@@ -7,7 +7,7 @@ use crate::byte_chunk::{ByteChunk, Chunk};
 /// Runs in O(N) time.
 #[inline]
 pub fn count(text: &str) -> usize {
-    count_chars_internal::<Chunk>(text.as_bytes())
+    count_internal::<Chunk>(text.as_bytes())
 }
 
 /// Converts from byte-index to char-index in a string slice.
@@ -20,14 +20,11 @@ pub fn count(text: &str) -> usize {
 /// Runs in O(N) time.
 #[inline]
 pub fn from_byte_idx(text: &str, byte_idx: usize) -> usize {
-    let count = count_chars_internal::<Chunk>(
-        &text.as_bytes()[0..(byte_idx.saturating_add(1)).min(text.len())],
-    );
-    if byte_idx < text.len() {
-        count - 1
-    } else {
-        count
+    let mut i = byte_idx.min(text.len());
+    while !text.is_char_boundary(i) {
+        i -= 1;
     }
+    count_internal::<Chunk>(&text.as_bytes()[0..i])
 }
 
 /// Converts from char-index to byte-index in a string slice.
@@ -54,50 +51,61 @@ fn to_byte_idx_inner<T: ByteChunk>(text: &str, char_idx: usize) -> usize {
     let mut char_count = 0;
 
     // Take care of any unaligned bytes at the beginning.
-    let mut i = 0;
-    while i < start.len() && char_count <= char_idx {
-        char_count += ((start[i] & 0xC0) != 0x80) as usize;
-        i += 1;
-    }
-    byte_count += i;
-
-    // Use chunks to count multiple bytes at once.
-    let mut i = 0;
-    let mut acc = T::zero();
-    let mut acc_i = 0;
-    while i < middle.len() && (char_count + (T::size() * (acc_i + 1))) <= char_idx {
-        acc = acc.add(middle[i].bitand(T::splat(0xc0)).cmp_eq_byte(0x80));
-        acc_i += 1;
-        // if acc_i == T::max_acc() || (char_count + (T::size() * (acc_i + 1))) >= char_idx {
-        if acc_i == T::max_acc() || (char_count + (T::size() * (acc_i + 1))) >= char_idx {
-            char_count += (T::size() * acc_i) - acc.sum_bytes();
-            acc_i = 0;
-            acc = T::zero();
+    for byte in start.iter() {
+        char_count += ((*byte & 0xC0) != 0x80) as usize;
+        if char_count > char_idx {
+            break;
         }
-        i += 1;
+        byte_count += 1;
+    }
+
+    // Process chunks in the fast path.
+    let mut chunks = middle;
+    let mut max_round_len = char_idx.saturating_sub(char_count) / T::MAX_ACC;
+    while max_round_len > 0 && !chunks.is_empty() {
+        // Choose the largest number of chunks we can do this round
+        // that will neither overflow `max_acc` nor blast past the
+        // remaining line breaks we're looking for.
+        let round_len = T::MAX_ACC.min(max_round_len).min(chunks.len());
+        max_round_len -= round_len;
+        let round = &chunks[..round_len];
+        chunks = &chunks[round_len..];
+
+        // Process the chunks in this round.
+        let mut acc = T::zero();
+        for chunk in round.iter() {
+            acc = acc.add(chunk.bitand(T::splat(0xc0)).cmp_eq_byte(0x80));
+        }
+        char_count += (T::SIZE * round_len) - acc.sum_bytes();
+        byte_count += T::SIZE * round_len;
+    }
+
+    // Process chunks in the slow path.
+    for chunk in chunks.iter() {
+        let new_char_count =
+            char_count + T::SIZE - chunk.bitand(T::splat(0xc0)).cmp_eq_byte(0x80).sum_bytes();
+        if new_char_count >= char_idx {
+            break;
+        }
+        char_count = new_char_count;
+        byte_count += T::SIZE;
     }
-    char_count += (T::size() * acc_i) - acc.sum_bytes();
-    byte_count += i * T::size();
 
     // Take care of any unaligned bytes at the end.
     let end = &text.as_bytes()[byte_count..];
-    let mut i = 0;
-    while i < end.len() && char_count <= char_idx {
-        char_count += ((end[i] & 0xC0) != 0x80) as usize;
-        i += 1;
+    for byte in end.iter() {
+        char_count += ((*byte & 0xC0) != 0x80) as usize;
+        if char_count > char_idx {
+            break;
+        }
+        byte_count += 1;
     }
-    byte_count += i;
 
-    // Finish up
-    if byte_count == text.len() && char_count <= char_idx {
-        byte_count
-    } else {
-        byte_count - 1
-    }
+    byte_count
 }
 
 #[inline(always)]
-fn count_chars_internal<T: ByteChunk>(text: &[u8]) -> usize {
+pub(crate) fn count_internal<T: ByteChunk>(text: &[u8]) -> usize {
     // Get `middle` for more efficient chunk-based counting.
     let (start, middle, end) = unsafe { text.align_to::<T>() };
 
@@ -109,19 +117,14 @@ fn count_chars_internal<T: ByteChunk>(text: &[u8]) -> usize {
     }
 
     // Take care of the middle bytes in big chunks.
-    let mut i = 0;
-    let mut acc = T::zero();
-    for chunk in middle.iter() {
-        let tmp = chunk.bitand(T::splat(0xc0)).cmp_eq_byte(0x80);
-        acc = acc.add(tmp);
-        i += 1;
-        if i == T::max_acc() {
-            i = 0;
-            inv_count += acc.sum_bytes();
-            acc = T::zero();
+    for chunks in middle.chunks(T::MAX_ACC) {
+        let mut acc = T::zero();
+        for chunk in chunks.iter() {
+            let tmp = chunk.bitand(T::splat(0xc0)).cmp_eq_byte(0x80);
+            acc = acc.add(tmp);
         }
+        inv_count += acc.sum_bytes();
     }
-    inv_count += acc.sum_bytes();
 
     // Take care of unaligned bytes at the end.
     for byte in end.iter() {
